/**************************************************************************
/* ShadowJAAS - a JAAS authentication provider for the
/* shadow-password system of Linux.
/*
/* Copyright (c) 2002 by Bernhard Bablok (mail@bablokb.de)
/*
/* This program is free software; you can redistribute it and/or modify
/* it under the terms of the GNU Library General Public License as published
/* by  the Free Software Foundation; either version 2 of the License or
/* (at your option) any later version.
/*
/* This program is distributed in the hope that it will be useful, but
/* WITHOUT ANY WARRANTY; without even the implied warranty of
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/* GNU Library General Public License for more details.
/*
/* You should have received a copy of the GNU Library General Public License
/* along with this program; see the file COPYING.LIB.  If not, write to
/* the Free Software Foundation Inc., 59 Temple Place - Suite 330,
/* Boston, MA  02111-1307 USA
/**************************************************************************/

package de.bablokb.jaas.auth.module;

import java.security.*;

import java.util.*;
import java.io.*;
import javax.security.auth.*;
import javax.security.auth.callback.*;
import javax.security.auth.login.*;
import javax.security.auth.spi.*;

import com.sun.security.auth.*;   /* principal classes */

/**
   This LoginModule uses the files /etc/{passwd,shadow} to authenticate
   the user. Since /etc/shadow is typically accessible only by root,
   the module uses a native program called 'shadowjaas' which should run
   suid root (and should be in the PATH).

   <p> After authentication, a number of principals are added to the
   subject: UnixPrincipal (user name), UnixNumericUserPrincipal (uid)
   UnixNumericGroupPrincipals (GIDs).

   <p> This LoginModule recognizes the debug option.
   If set to true in the login Configuration,
   debug messages will be output to the output stream, System.out.
  
   @version $Revision: 1.5 $
   @author  $Author: Bablokb $
*/

public class ShadowLoginModule implements LoginModule {

  private final static int 
    DEBUG_USERNAME=0, 
    DEBUG_PASSWORD=1,
    DEBUG_AUTH=2,
    DEBUG_NOAUTH=3,
    DEBUG_ADD_PRINCIPALS=4,
    DEBUG_REMOVE_PRINCIPALS=5,
    DEBUG_DEFAULT=1000;

  private final static String NATIVE_CMD = "shadowjaas";

  private Subject iSubject;
  private CallbackHandler iCallbackHandler;

  private boolean iDebug       = false;
  private boolean iDebugNative = false;
  private boolean iAuthOk      = false;
  private boolean iCommitOk    = false;

  private String  iUserName;
  private char[]  iPassword;
  private HashSet iPrincipals, iGroups;
  private boolean iPrimaryGroup;

  ////////////////////////////////////////////////////////////////////////////

  /**
     Initialize this <code>LoginModule</code>. Implements
     <code>javax.security.auth.spi.LoginModule.initialize(...)</code>.
    
     @param subject the <code>Subject</code> to be authenticated.
    
     @param callbackHandler a <code>CallbackHandler</code> for communicating
     with the end user (prompting for usernames and passwords, for example).
    
     @param sharedState state shared witch other configured 
     <code>LoginModules</code> (unused).
    
     @param options options specified in the login
     <code>Configuration</code> for this particular <code>LoginModule</code>
     (options honoured: debug and debug-native).
   */

  public void initialize(Subject subject, CallbackHandler callbackHandler,
			 Map sharedState, Map options) {

    iSubject = subject;
    iCallbackHandler = callbackHandler;
    iDebug = "true".equalsIgnoreCase((String)options.get("debug"));
    iDebugNative = "true".equalsIgnoreCase((String)options.get("debug-native"));
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Method to authenticate a subject (phase 1). Implements
     <code>javax.security.auth.spi.LoginModule.login()</code>.    
    
     @exception FailedLoginException if the authentication fails.
    
     @exception LoginException if this <code>LoginModule</code>
     is unable to perform the authentication.

     @return true in all cases since this <code>LoginModule</code>
     should not be ignored.
  */

  public boolean login() throws LoginException {
    if (iCallbackHandler == null)
      throw new LoginException("error: no CallbackHandler available");

    Callback[] cb = new Callback[2];
    cb[0] = new NameCallback("username: ");
    cb[1] = new PasswordCallback("password: ",false);

    try {
      iCallbackHandler.handle(cb);
      iUserName = ((NameCallback)cb[0]).getName();

      char[] tmpPassword = ((PasswordCallback)cb[1]).getPassword();
      if (tmpPassword == null)
	tmpPassword = new char[0];
      iPassword = new char[tmpPassword.length];
      System.arraycopy(tmpPassword,0,iPassword,0,tmpPassword.length);
      ((PasswordCallback)cb[1]).clearPassword();

    } catch (IOException ioe) {
      throw new LoginException(ioe.toString());
    } catch (UnsupportedCallbackException uce) {
      throw new LoginException(uce.getCallback().toString());
    }

    printDebug(DEBUG_USERNAME);
    printDebug(DEBUG_PASSWORD);

    if (verifyPassword()) {
      printDebug(DEBUG_AUTH);
      iAuthOk = true;
      return true;
    } else {  
      printDebug(DEBUG_NOAUTH);
      clearState();
      throw new FailedLoginException();
    }
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Method to commit the authentication process (phase 2). Implements
     <code>javax.security.auth.spi.LoginModule.commit()</code>.    

     <p> This method is called if the LoginContext's
     overall authentication succeeded
     (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
     succeeded).
    
     <p> If this LoginModule's own authentication attempt
     succeeded (checked by retrieving the private state saved by the
     <code>login</code> method), then this method associates relevant
     Principals and Credentials with the <code>Subject</code> located in the
     <code>LoginModule</code>.  If this LoginModule's own
     authentication attempted failed, then this method removes
     any state that was originally saved.
    
     @exception LoginException if the commit fails.
    
     @return true if this LoginModule's own login and commit
     attempts succeeded, or false otherwise.
   */

  public boolean commit() throws LoginException {
    if (iAuthOk == false) {
      return false;
    } else {
      addPrincipals();
      clearState();
      iCommitOk = true;
      return true;
    }
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Method to abort the authentication process (phase 2). Implements
     <code>javax.security.auth.spi.LoginModule.abort()</code>.    

     <p> This method is called if the LoginContext's
     overall authentication failed.
     (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
     did not succeed).
    
     <p> If this LoginModule's own authentication attempt
     succeeded (checked by retrieving the private state saved by the
     <code>login</code> and <code>commit</code> methods),
     then this method cleans up any state that was originally saved.
    
     @exception LoginException if the abort fails.
    
     @return false if this LoginModule's own login and/or commit attempts
     failed, and true otherwise.
   */

  public boolean abort() throws LoginException {
    if (iAuthOk == false) {
      return false;
    } else if (iAuthOk == true && iCommitOk == false) {
      iAuthOk = false;
      clearState();
    } else {
      logout();  // someone else failed
    }
    return true;
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Method which logs out a Subject. Implements
     <code>javax.security.auth.spi.LoginModule.logout()</code>.    

     <p> This method removes the Principals
     that were added by the <code>commit</code> method.
    
    
     @exception LoginException if the logout fails.
    
     @return true in all cases since this <code>LoginModule</code>
     should not be ignored.
   */

  public boolean logout() throws LoginException {
    removePrincipals();
    iAuthOk = iCommitOk = false;
    clearState();
    return true;
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Verify password. This method uses the native program shadowjaas.

     @return true if username is authenticated by the shadow password system
  */

  private boolean verifyPassword() throws LoginException {
    int rc = -1;
    try {
      String cmd;
      if (iDebugNative)
	cmd = NATIVE_CMD+" -d";
      else
	cmd = NATIVE_CMD;
      Process pr = Runtime.getRuntime().exec(cmd);
      
      // write username/password to stdin of native program

      PrintWriter writer = new PrintWriter(pr.getOutputStream());
      writer.println(iUserName);
      if (iPassword.length > 0)
	writer.println(iPassword);
      writer.flush();
      writer.close();
      writer = null;

      pr.waitFor();
      rc = pr.exitValue();
      if (rc == 0) {
	// add UnixPrincipal
	iPrincipals = new HashSet();
	iGroups     = new HashSet();
	iPrimaryGroup = true;
	UnixPrincipal up = new UnixPrincipal(iUserName);
	iPrincipals.add(up);

	//add UnixNumericUserPrincipal/UnixNumericGroupPrincipal
	InputStreamReader isr = new InputStreamReader(pr.getInputStream());
	BufferedReader input  = new BufferedReader(isr);
	while (true) {
	  String line = input.readLine();
	  if (line == null)
	    break;
	  else
	    addPrincipal(line);
	}
	input.close();
      }
      if (iDebugNative) {
	InputStreamReader isr = new InputStreamReader(pr.getErrorStream());
	BufferedReader input  = new BufferedReader(isr);
	while (true) {
	  String line = input.readLine();
	  if (line == null)
	    break;
	  else
	    System.err.println(line);
	}
	input.close();
      }
    } catch (Exception e) {
      throw new LoginException(e.toString());
    }
    return rc == 0;
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Add a principal to the internal set of principals. The internal
     set is guaranteed to exist.

     <p>N.B.: since a username might or might not be listed in the
     primary group in /etc/groups, we have to make sure not to add a group
     twice. Since <code>UnixNumericGroupPrincipal.equals()</code> also
     takes the primary-group flag into account, we cannot rely on the
     iPrincipals-Set to only add a group once.
  */

  private void addPrincipal(String line) {
    StringTokenizer t = new StringTokenizer(line," ");
    if (t.countTokens() == 1)
      iPrincipals.add(new UnixNumericUserPrincipal(t.nextToken()));
    else {
      String gid = t.nextToken();
      if (!iGroups.contains(gid)) {
	iGroups.add(gid);
	iPrincipals.add(new UnixNumericGroupPrincipal(gid,iPrimaryGroup));
//      iPrincipals.add(new UnixGroupPrincipal(t.nextToken(),
//                                           iPrimaryGroup));
      iPrimaryGroup = false;
      }
    }
  }


  ////////////////////////////////////////////////////////////////////////////

  /**
     Add principals to the subject.
  */

  private void addPrincipals() {
    if (iPrincipals == null || iPrincipals.size() == 0)
      return;
    AccessController.doPrivileged(
      new PrivilegedAction() {
	public Object run() {
	  Set currentPrincipals = iSubject.getPrincipals();
	  Iterator it = iPrincipals.iterator();
	  while (it.hasNext()) {
	    Principal p = (Principal) it.next();
	    if (!currentPrincipals.contains(p))
	      currentPrincipals.add(p);
	  }
	  return null;
	}
      }
    );
    printDebug(DEBUG_ADD_PRINCIPALS);
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Remove principals.
  */

  private void removePrincipals() {
    if (iPrincipals == null || iPrincipals.size() == 0)
      return;
    AccessController.doPrivileged(
      new PrivilegedAction() {
	public Object run() {
	  Set currentPrincipals = iSubject.getPrincipals();
	  Iterator it = iPrincipals.iterator();
	  while (it.hasNext()) {
	    Principal p = (Principal) it.next();
	    if (currentPrincipals.contains(p))
	      currentPrincipals.remove(p);
	  }
	  return null;
	}
      }
    );
    printDebug(DEBUG_REMOVE_PRINCIPALS);
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Clear state.
  */

  private void clearState() {
    iUserName = null;
    if (iPrincipals != null)
      iPrincipals.clear();
    iPrincipals = null;
    clearPassword();
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Clear password information.
  */

  private void clearPassword() {
    if (iPassword != null) {
      for (int i = 0; i < iPassword.length; i++)
	iPassword[i] = ' ';
      iPassword = null;
    }
  }

  ////////////////////////////////////////////////////////////////////////////

  /**
     Print debug information.
  */

  private void printDebug(int debugType) {
    if (!iDebug)
      return;
    switch (debugType) {
    case DEBUG_USERNAME:
      System.err.println("ShadowLoginModule: username="+iUserName);
      break;
    case DEBUG_PASSWORD:
      System.err.print("ShadowLoginModule: password=");
      for (int i = 0; i < iPassword.length; i++)
	System.err.print(iPassword[i]);
      System.err.println();
      break;
    case DEBUG_AUTH:
	System.err.println("ShadowLoginModule: authentication succeeded");
	break;
    case DEBUG_NOAUTH:
	System.err.println("ShadowLoginModule: authentication failed");
	break;
    case DEBUG_ADD_PRINCIPALS:
	System.err.println("ShadowLoginModule: added principals");
	break;
    case DEBUG_REMOVE_PRINCIPALS:
	System.err.println("ShadowLoginModule: removed principals");
	break;
    default:
      break;
    }
  }
}
