/**************************************************************************
/* ShadowJAAS - a JAAS authentication provider for the
/* shadow-password system of Linux.
/*
/* Copyright (c) 2002 by Bernhard Bablok (mail@bablokb.de)
/*
/* This program is free software; you can redistribute it and/or modify
/* it under the terms of the GNU Library General Public License as published
/* by  the Free Software Foundation; either version 2 of the License or
/* (at your option) any later version.
/*
/* This program is distributed in the hope that it will be useful, but
/* WITHOUT ANY WARRANTY; without even the implied warranty of
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/* GNU Library General Public License for more details.
/*
/* You should have received a copy of the GNU Library General Public License
/* along with this program; see the file COPYING.LIB.  If not, write to
/* the Free Software Foundation Inc., 59 Temple Place - Suite 330,
/* Boston, MA  02111-1307 USA
/**************************************************************************/

/* This program takes no arguments:
     shadowjaas [-d]
   and reads username and password from stdin
   and checks the password against /etc/{passwd,shadow}

   The program returns 0 in case the password is ok, 1 otherwise.
   The option -d outputs debug-messages.

   $Revision: 1.4 $
   $Author: Bablokb $
*/   

#define _XOPEN_SOURCE
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <shadow.h>
#include <stdio.h>
#include <stdarg.h>
#include <search.h>

#define MAX_USERNAME_SIZE 32
#define MAX_PASSWORD_SIZE 28

void parseOptions(int argc, char **argv);
void readPasswords(void);
void usage(const char *);
void writeMsg(const char *msg, ...);
int  checkPassword(void);
char *encryptPassword(void);
void dumpIds(void);

const int OK = 1;
const int FAILED = 0;

int  doDebug = 0;
char userName[MAX_USERNAME_SIZE];
char clearTextPassword[MAX_PASSWORD_SIZE];
char *encryptedPassword;

uid_t userId;
gid_t groupId;

int main(int argc, char **argv) {
  parseOptions(argc,argv);
  readPasswords();
  if (checkPassword() == OK) {
    writeMsg("password check ok\n");
    dumpIds();
    exit(0);
  } else {
    writeMsg("password check failed\n");
    exit(1);
  }
}

/* parse arguments   ------------------------------------------------------ */
 
void parseOptions(int argc, char **argv) {
  int option;
  char format[8];

  while ((option = getopt(argc, argv, "d")) != EOF) {
    switch (option) {
    case 'd':
      doDebug = 1;
      break;
    default:
      usage(argv[0]);
    }
  }

  userName[0] = '\0';
  if (!feof(stdin)) {
    sprintf(format,"%%%ds",MAX_USERNAME_SIZE);
    fscanf(stdin,format,userName);
    userName[sizeof(userName)-1] = '\0';
    writeMsg("username: %s\n",userName);
  } else
    usage(argv[0]);
  
  clearTextPassword[0] = '\0';
  if (!feof(stdin)) {
    sprintf(format,"%%%ds",MAX_PASSWORD_SIZE);
    fscanf(stdin,format,clearTextPassword);
    clearTextPassword[sizeof(clearTextPassword)-1] = '\0';
  } /* else: empty password */
  writeMsg("cleartext password: %s\n",clearTextPassword);
}
   
/* read information from /etc/{passwd,shadow}   --------------------------- */

void readPasswords() {
  struct passwd *pwd;
  struct spwd *spwd = NULL;
  if (!(pwd = getpwnam(userName))) {
    writeMsg("no entry for %s in /etc/passwd\n",userName);
    exit(3);
  } else {
    encryptedPassword = pwd->pw_passwd;
    userId = pwd->pw_uid;
    groupId = pwd->pw_gid;
  }
  endpwent();

  if (strcmp(encryptedPassword,"x") == 0 || strcmp(encryptedPassword,"*") == 0) {
    if (!(spwd = getspnam(userName))) {
      writeMsg("no entry for %s in /etc/shadow\n",userName);
      exit(3);
    } else
      encryptedPassword = spwd->sp_pwdp;
    endspent();
  }
  writeMsg("encrypted password: %s\n",encryptedPassword);
  writeMsg("numeric user-id: %lu\n",userId);
  writeMsg("numeric group-id: %lu\n",groupId);
}

/* check password   ------------------------------------------------------- */

int checkPassword() {
  int rc;

  if (*clearTextPassword == '\0' && (
        encryptedPassword == (char *) 0 || *encryptedPassword == '\0'))
    return OK;
  else if (*clearTextPassword != '\0' && (
	    encryptedPassword == (char *) 0 || *encryptedPassword == '\0')) {
    bzero(clearTextPassword,strlen(clearTextPassword));
    return FAILED;
  }

  rc = strcmp(encryptPassword(),encryptedPassword);
  bzero(clearTextPassword,strlen(clearTextPassword));
  if (rc == 0)
    return OK;
  else
    return FAILED;
}

/* encrypt cleartext password   ------------------------------------------- */

char *encryptPassword() {
  static char cipher[128];
  char *p;
  const char *salt = encryptedPassword;
  p = crypt(clearTextPassword,salt);
  if (strlen(p) > sizeof(cipher)-1)
    p[sizeof(cipher)-1] = 0;
  strcpy(cipher,p);
  writeMsg("encrypted password: %s\n",cipher);
  return cipher;
}

/* dump user/group ids   ------------------------------------------------- */

void dumpIds() {
  struct group *gp;
  char **cp;
  printf("%lu\n%lu",userId,groupId);
  if (gp = getgrgid(groupId))
    printf(" %s\n",gp->gr_name);
  else
    printf("\n");
  while (gp = getgrent()) {
    cp = gp->gr_mem;
    while (*cp) {
      if (!strcmp(userName,*cp)) {
	  printf("%lu %s\n",gp->gr_gid,gp->gr_name);
	  break;
      }
      cp++;
    }
  }
  endgrent();
}

/* show usage information   ---------------------------------------------- */

void usage(const char *pgmName) {
  printf("usage: %s [-d] username password\n",pgmName);
  exit(3);
}

/* write debug message   ------------------------------------------------- */

void writeMsg(const char *msg, ...) {
  if (doDebug) {
    va_list p;
    fprintf(stderr,"shadowjaas: ");
    va_start(p,msg);
    fflush(stdout);
    vfprintf(stderr,msg,p);
    fflush(stderr);
    va_end(p);
  }
}
